{{
****************************************
* File jvmException: exception handler *
****************************************
}}

{{
  Exception Handling.  
  Because exception handling is a major feature of the Java programming language, you might be surprised 
  to learn that exceptions are handled with a single bytecode instruction. When exceptions occur, the object 
  thrown must be a reference to an object of the subclass Throwable, and the current Java stackframe is 
  searched for the most recent catch clause that catches exceptions of this class or a superclass of this 
  class. When a matching catch clause is found, the program counter is reset to the address indicated by 
  the catch clause and execution continues from there. When no appropriate catch clause is found, that 
  frame is popped and the object is rethrown. If a catch clause is then found, the clause will contain the 
  location of the code for the exception, the program counter is reset to that location, and execution 
  continues. Otherwise, the frame is popped and the object is rethrown. 
  Note:
  A NullPointerException is thrown instead if the referenced object is null.  
}}

{{
  athrow
  Operation: Throw exception or error
  Format: athrow   
  Operand Stack: ...,objectref --> objectref

  Description
  The objectref must be of type reference and must refer to an object that is an instance of class Throwable
  or of a subclass of Throwable. It is popped from the operand stack. The objectref is then thrown by searching
  the current method for the first exception handler that matches the class of objectref.
  If an exception handler that matches objectref is found, it contains the location of the code intended to
  handle this exception. The pc register is reset to that location, the operand stack of the current frame is
  cleared, objectref is pushed back onto the operand stack, and execution continues. 
  If no matching exception handler is found in the current frame, that frame is popped. If the current frame
  represents an invocation of a synchronized method, the monitor acquired or reentered on invocation of the
  method is released or exited (respectively) as if by execution of a monitorexit instruction. Finally, the frame
  of its invoker is reinstated, if such a frame exists, and the objectref is rethrown. If no such frame exists,
  the current thread exits.
  Runtime Exceptions
  If objectref is null, athrow throws a NullPointerException instead of objectref.
  Notes
  The operand stack diagram for the athrow instruction may be misleading: If a handler for this exception is
  matched in the current method, the athrow instruction discards all the values on the operand stack, then pushes
  the thrown object onto the operand stack. However, if no handler is matched in the current method and the
  exception is thrown farther up the method invocation chain, then the operand stack of the method (if any) that
  handles the exception is cleared and objectref is pushed onto that empty operand stack. All intervening frames
  from the method that threw the exception up to, but not including, the method that handles the exception are
  discarded.
}}


CON
  'exceptions
  kIllegalArgumentException  = 1
  kNullPointerException      = 2
  kIndexOutOfBoundsException = 3
  kOutOfMemoryError          = 4


OBJ
  js:   "jvmStack"
  jd:   "jvmData"
  comm: "jvmComm"
  jsub: "jvmSubReturn"

  
PUB jb_athrow | aref,class
{{ Meaning: Throw exception or error. }}
{{
  The objectref must be of type reference and must refer to an object that is an instance of class Throwable or
  of a subclass of Throwable. It is popped from the operand stack. The objectref is then thrown by searching the
  current method for the first exception handler that matches the class of objectref.
  If an exception handler that matches objectref is found, it contains the location of the code intended to handle
  this exception. The pc register is reset to that location, the operand stack of the current frame is cleared,
  objectref is pushed back onto the operand stack, and execution continues. 
  If no matching exception handler is found in the current frame, that frame is popped. If the current frame
  represents an invocation of a synchronized method, the monitor acquired or reentered on invocation of the method
  is released or exited (respectively) as if by execution of a monitorexit instruction.
  Finally, the frame of its invoker is reinstated, if such a frame exists, and the objectref is rethrown.
  If no such frame exists, the current thread exits.
}}
{{
  JEM format: athrow
}}
  aref := js.pop 'get objectref
  throwException(aref)

    
PUB throw(exceptionNumber) | method
{{ Auxiliary function (no btecode): throw one of four runtime exceptions }}
  method := jd.readJavaIntBE(jd#JVM_THROW)
  jd.writeJavaIntBE(jd.getJvmFP,exceptionNumber)        'write exceptionNumber to arg0 in current frame (eg. istore_0)
                                                        'because method throwVMException starts with iload_0 
  jd.setJvmEP(jd.getJvmPC)                              'remember where exception is thrown
  jd.setJvmPC(jd.readJavaIntBE(method + jd#JVM_CODE))   'set new pc to method throwVMException

  
PUB throwException(e) | exceptionTable,numExceptions,index,paramsize,localsize,pc,sp,class
{{ Auxiliary function (no bytecode): Throw an exception }}
  if e <> 0
    if jd.getJvmEP == 0
      jd.setJvmEP(jd.getJvmPC) 'set exception if not already set
    repeat
      exceptionTable := jd.readJavaIntBE(jd.getJvmMP + jd#JVM_EXCEPTION) 'get address of exception table
      numExceptions := 0
      if exceptionTable <> 0
        numExceptions := jd.readJavaByte(exceptionTable) 'read number of exception handlers
      index := 0
      repeat while numExceptions > 0
        if jd.getJvmPC => jd.readJavaIntBE(exceptionTable + 1 + index) AND jd.getJvmPC < jd.readJavaIntBE(exceptionTable + 3 + index)
          if instanceOf(e,jd.readJavaIntBE(exceptionTable + 5 + index))
            'clear operand stack
            paramsize := (jd.readJavaByte(jd.getJvmMP + jd#JVM_PARAMETERS) & $7F) 'mask out exception bit
            localsize := jd.readJavaByte(jd.getJvmMP + jd#JVM_LOCALS)
            sp := jd.getJvmFP + paramsize + localsize + 6
            jd.setJvmSP(sp)
            'push exception
            js.push(e)
            'jump to handler
            pc := jd.readJavaIntBE(exceptionTable + 7 + index)
            jd.setJvmPC(pc)
            return
        index += 8 'size of exception record
        numExceptions--
      jsub.jb_return  'pop frame
    until jd.getJvmMP == $FFFF
    class := jd.readJavaIntBE(e)
    comm.errException(class,jd.getJvmEP) 'signal exception to IDE
    jd.setJvmEP(0) 'clear exception
  else
    throw(kNullPointerException)
    
      
PUB instanceOf(instance,class): YesNo | instanceclass
{{ Auxiliary function (no bytecode) Meaning: Determine if an object is of given type and return result. }}
{{
  Description: Determines whether the referenced object can be cast to be a reference to an 
  object of the expected class. It returns true or false.
}}
  if instance <> 0
    instanceclass := jd.readJavaIntBE(instance) 'get class type for instance
    repeat
      if instanceclass == class
        return true
      else
        instanceclass := jd.readJavaIntBE(instanceclass) 'get superclass of instance
    until instanceclass == $FFFF '$FFFF is superclass of Object class
  return false
  
    